from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto

from isbnlib import Isbn, meta
from isbnlib._exceptions import NotValidISBNError

__all__ = ["validate_isbn", "get_isbn_metadata", "BookMetadata"]


def validate_isbn(isbn_code: str) -> bool:
    try:
        Isbn(isbn_code)
    except NotValidISBNError:
        is_isbn = False
    else:
        is_isbn = True
    return is_isbn


class TitledEnum(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name.title()

    @classmethod
    def names2values(cls) -> list[str]:
        return {m.name: v for v, m in cls._value2member_map_.items()}


class MetaEnum(TitledEnum):
    """
    Internal representations of keys from the :class:`isbnlib.Isbn` class.
    """

    title = auto()
    authors = auto()
    year = auto()
    isbn_13 = "ISBN-13"
    publisher = auto()


def get_isbn_metadata(isbn_code: str) -> dict[str, str | list[str]]:
    book_metadata = meta(isbn_code)
    # title, authors, year, isbn_13, publisher = map(book_metadata.get, MetaEnum.values())
    book_meta_kwargs = {
        name: book_metadata.get(value)
        for name, value in MetaEnum.names2values().items()
    }
    meta_cls = BookMetadata(**book_meta_kwargs)
    return meta_cls


class IndexedBook:
    @property
    def first_author(self):
        return self.authors[0]

    @property
    def first_author_surname(self) -> str:
        return self.author_surname(self.first_author)

    @staticmethod
    def author_surname(name: str) -> str:
        return name.split(" ")[-1]

    @property
    def short_fmt(self) -> str:
        return f"{self.first_author_surname} ({self.year}) {self.title}"


@dataclass
class BookMetadata(IndexedBook):
    title: str
    authors: list[str]
    year: int
    publisher: str
    isbn_13: str

    def __repr__(self) -> str:
        return f"'\N{open book}: {self.short_fmt}'"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from dataclass_wizard import JSONWizard

__all__ = ["DocTreeDoc", "Page", "Block", "Line", "Word"]


@dataclass
class DocTreeDoc(JSONWizard):
    pages: list[Page] = field(default_factory=list)

    class _(JSONWizard.Meta):
        # Sets the target key transform to use for serialization;
        # defaults to `camelCase` if not specified.
        key_transform_with_dump = "SNAKE"


@dataclass
class Page:
    page_idx: int
    dimensions: tuple[int, int]
    # orientation: dict[str, Any]
    # language: dict[str, Any]
    blocks: list[Block]


@dataclass
class Block:
    # geometry: tuple[tuple[float, float], tuple[float, float]]
    lines: list[Line]


@dataclass
class Line:
    # geometry: tuple[tuple[float, float], tuple[float, float]]
    words: list[Word]


@dataclass
class Word:
    value: str
    confidence: float
    # geometry: tuple[tuple[float, float], tuple[float, float]]



r"""
:mod:`dex` is an index under construction"""

from .library import load_library

__all__ = ["library"]

__author__ = "Louis Maddox"
__license__ = "MIT"
__description__ = "Indexer under construction."
__url__ = "https://github.com/lmmx/dex"
__uri__ = __url__
__email__ = "louismmx@gmail.com"



import multiprocessing as mp
from multiprocessing import Pool, Process

from more_itertools import chunked
from tqdm import tqdm

__all__ = ["batch_multiprocess", "batch_multiprocess_with_return"]


def batch_multiprocess(
    function_list, n_cores=mp.cpu_count(), show_progress=True, tqdm_desc=None
):
    """
    Run a list of functions on `n_cores` (default: all CPU cores),
    with the option to show a progress bar using tqdm (default: shown).
    """
    iterator = [*chunked(function_list, n_cores)]
    if show_progress:
        iterator = tqdm(iterator, desc=tqdm_desc)
    for func_batch in iterator:
        procs = []
        for f in func_batch:
            procs.append(Process(target=f))
        for p in procs:
            p.start()
        for p in procs:
            p.join()


def batch_multiprocess_with_return(
    function_list,
    pool_results=None,
    n_cores=mp.cpu_count(),
    show_progress=True,
    tqdm_desc=None,
):
    """
    Run a list of functions on `n_cores` (default: all CPU cores),
    with the option to show a progress bar using tqdm (default: shown).
    """
    iterator = [*chunked(function_list, n_cores)]
    pool_results = pool_results if pool_results else []
    pool = Pool(processes=n_cores)
    if show_progress:
        iterator = tqdm(iterator, desc=tqdm_desc)
    for func_batch in iterator:
        for f in func_batch:
            pool.apply_async(func=f, callback=pool_results.append)
    pool.close()
    pool.join()
    return pool_results



from __future__ import annotations

import logging


class Console:
    def __init__(self, name=None, level=logging.INFO, timestamped=False):
        self.logger = logging.getLogger(name=name)
        self.console = logging.StreamHandler()
        for target in (self.logger, self.console):
            target.setLevel(level)
        self.timestamped = timestamped
        self.console.setFormatter(self.log_format)
        # If you get doubled up logs in your application, comment out the next line:
        self.logger.addHandler(self.console)

    @property
    def header_names(self) -> list[str]:
        """The log format header/s (with/without timestamp)."""
        return [*filter(lambda t: self.timestamped, ["asctime"]), "levelname"]

    @property
    def log_format(self) -> logging.Formatter:
        header_str = " ".join(map("[%({})s]".format, self.header_names))
        return logging.Formatter(fmt=header_str + " %(message)s")



from __future__ import annotations

import re
from dataclasses import dataclass
from functools import partial
from pathlib import Path

from .isbn_utils import BookMetadata, get_isbn_metadata
from .log_utils import Console
from .multiproc_utils import batch_multiprocess_with_return
from .ocr_utils import scan_text_in_images
from .path_utils import shelves_path

__all__ = ["load_library"]

logger = Console(name=__name__).logger


def load_library() -> Library:
    pattern = "dex*.zip"
    zipfiles = [*shelves_path.glob(pattern)]
    lib = Library.from_zips(zipfiles)
    return lib


class Library(list):
    def __init__(self, library_items: list[LibraryItem]):
        if not library_items:
            logger.info("Library is empty")
        else:
            logger.info(f"Library contains {len(library_items)} files")
        self.append(library_items)

    @classmethod
    def from_zips(cls, zipfiles: list[Path], parallel: bool = False):
        if parallel:
            zip_items = batch_multiprocess_with_return(
                function_list=[partial(LibraryItem.from_zip, zf) for zf in zipfiles]
            )
        else:
            zip_items = [LibraryItem.from_zip(zf) for zf in zipfiles]
        return cls(library_items=zip_items)


class IndexedItem:
    isbn_regex = r"[0-9]{10,13}"

    @property
    def is_archived(self) -> bool:
        return self.archive is not None

    @classmethod
    def from_isbn(cls, isbn_code: str, archive_path: Path | None = None) -> LibraryItem:
        metadata = get_isbn_metadata(isbn_code)
        return cls(metadata=metadata, archive_path=archive_path)

    @classmethod
    def from_zip(cls, zipfile: Path) -> LibraryItem:
        suffixes = [".zip"]
        if zipfile.suffix not in suffixes:
            raise ValueError(f"{zipfile} not recognised as a ZIP file")
        try:
            isbn_code = next(re.finditer(cls.isbn_regex, zipfile.stem)).group()
        except StopIteration:
            logger.error(f"Could not detect ISBN in filename {zipfile.stem} (omitting)")
        else:
            return cls.from_isbn(isbn_code=isbn_code, archive_path=zipfile)


@dataclass
class LibraryItem(IndexedItem):
    metadata: BookMetadata
    archive_path: Path | None



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path

from doctr.io import DocumentFile
from doctr.models import ocr_predictor

from .data_model import DocTreeDoc

__all__ = ["scan_text_in_images"]


def scan_text_in_images(image_paths: list[Path]):
    model_config = {
        "det_arch": "db_resnet50",
        "reco_arch": "crnn_vgg16_bn",
        "pretrained": True,
    }
    model = ocr_predictor(**model_config)
    multi_img_doc = DocumentFile.from_images(image_paths)
    result = model(multi_img_doc)
    result_dict = result.export()
    doc = DocTreeDoc.from_dict(result_dict)
    return doc



from __future__ import annotations

from os import environ
from pathlib import Path

from git import Repo
from git.exc import InvalidGitRepositoryError

from . import __path__

__all__ = ["shelves_path"]

pkg_path = Path(*__path__)


def _get_shelves_path() -> Path:
    """
    If the package is installed from the git repo, we can
    """
    env_var = "DEX_SHELVES"
    grandparent_dir = pkg_path.parent.parent

    if grandparent_dir.exists():
        try:
            Repo(grandparent_dir)
        except InvalidGitRepositoryError:
            is_git_repo = False
        else:
            is_git_repo = True
            shelves_path = grandparent_dir / "data" / "shelves"
    if not is_git_repo:
        if env_var in environ:
            shelves_path = Path(environ[env_var])
            if not shelves_path.exists():
                err_msg = f"{shelves_path=} set from {env_var=} does not exist"
                raise FileNotFoundError(err_msg)
        else:
            err_msg = (
                "Package data directory handling in distributed package not set up. "
                f"Please set a storage path as the '{env_var}' environment variable."
            )
            raise NotImplementedError(err_msg)
    return shelves_path


shelves_path = _get_shelves_path()

if not shelves_path.exists():
    raise FileNotFoundError("Internal error in library path resolution")
